#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include "giFTproto.h"
#include "giFTdownload.h"

/* The max number of sources to use in parallel to fetch this file. */
#define MAXSOURCES 200

/* How often between print status messages? */
#define MARK_INTERVAL 5

/**
 * Parse the given URL (which must be in the specific format:
 *   http://10.1.2.3:1214/4321/filename.ext
 * with a numeric IP address, and explicit port number) and add it to
 * the list of sources.
 */
static void addsource(DownloadContext dc, char *url)
{
    int res;
    unsigned int a1,a2,a3,a4,p;
    struct sockaddr_in sin;
    char *uri;

    if (strncmp(url, "http://", 7)) return;
    url += 7;

    res = sscanf(url, "%u.%u.%u.%u:%u", &a1, &a2, &a3, &a4, &p);
    if (res != 5) return;

    uri = strchr(url, '/');
    if (!uri) return;

    sin.sin_family = AF_INET;
    sin.sin_port = htons(p);
    sin.sin_addr.s_addr = htonl( (a1<<24) | (a2<<16) | (a3<<8) | a4 );

    downloadAddSource(dc, &sin, uri);
}

/**
 * Keep track of the number of active hosts we're downloading from.
 */
static int cchandler(int which, ConnectionState from, ConnectionState to,
	void *context)
{
    size_t *nhp = context;

    if (nhp == NULL) return -1;

    if (from == CONNECTED_DOWNLOAD_READING && to != CONNECTED_DOWNLOAD_READING
	    && (*nhp) > 0) {
	--(*nhp);
    }

    if (from != CONNECTED_DOWNLOAD_READING &&
	    to == CONNECTED_DOWNLOAD_READING) {
	++(*nhp);
    }

    return 0;
}

/**
 * Try to read URLs from the given fd.  When we find them, add them to
 * the list.  Return the new context.  If dc is NULL, free the given
 * context.
 */
static void *read_and_add_url(int fd, DownloadContext dc, void *context)
{
    size_t amt;
    int res;
    byte *start, *end;

    struct s_ctx {
	byte *data;
	size_t pos;
	size_t alloc;
    } *ctx = context;

    if (dc == NULL) {
	if (ctx != NULL) {
	    if (ctx->data) free(ctx->data);
	    free(ctx);
	}
	return NULL;
    } else if (ctx == NULL) {
	ctx = malloc(sizeof(struct s_ctx));
	if (ctx == NULL) return NULL;
	ctx->data = malloc(1025);
	if (ctx->data == NULL) {
	    free(ctx);
	    return NULL;
	}
	ctx->pos = 0;
	ctx->alloc = 1024;
    } else if (ctx->pos == ctx->alloc) {
	size_t newalloc = ctx->alloc + 1024;
	byte *newbuf = realloc(ctx->data, newalloc+1);
	if (newbuf == NULL) {
	    return ctx;
	}
	ctx->data = newbuf;
	ctx->alloc = newalloc;
    }

    amt = ctx->alloc - ctx->pos;
    res = read(fd, ctx->data + ctx->pos, amt);
    if (res <= 0) {
	free(ctx->data);
	free(ctx);
	return NULL;
    }
    ctx->pos += res;
    ctx->data[ctx->pos] = '\0';

    /* See if we have one or more complete URLs now. */

    while(1) {
	/* Skip anything that doesn't look like a URL. */
	start = strstr(ctx->data, "http://");
	if (!start) return ctx;
	/* Find the end of the URL. */
	end = start;
	while (*end && !strchr(" \r\n\"", *end)) ++end;
	if (*end == '\0') return ctx;
	*end = '\0';

	/* Add it. */
	addsource(dc, start);

	/* Remove it from the buffer. */
	memmove(ctx->data, end+1, ctx->data + ctx->pos - end);
	ctx->pos -= (end + 1 - ctx->data);
    }
}

/**
 * Save information about what parts of the file we already have.
 */
static int saveIntervals(char *filename, DownloadContext dc) {
    FILE *statef;
    void *ptr = NULL;
    size_t start, end;

    if((statef = fopen(filename, "w")) == NULL) return -1;
    do {
	ptr = downloadGetFilled(dc, ptr, &start, &end);
	if (start != end)
	    fprintf(statef, "%lu-%lu\n", (unsigned long)start,
		(unsigned long)end);
    } while (ptr);
    fclose(statef);
    return 0;
}

int main(int argc, char **argv)
{
    size_t filesize;
    FILE *outf, *statef;
    Connection *sources = NULL;
    void *rcontext = NULL;
    time_t marktime = 0;
    DownloadContext dc;
    int i;
    size_t nhosts = 0;
    int stdin_open = 1;
    void *dh_context = NULL;
    int success = 0;
    char *statefile;
    unsigned long start, end;
    size_t prev_recv = 0;
    char *open_mode;
    time_t now;

    /* Make sure stdout is never buffered, since we're likely to try to
     * use the status output as it is created. */
    setbuf(stdout, NULL);

    /* Check the arguments. */
    if (argc < 3 || (filesize = atoi(argv[1])) == 0) {
	fprintf(stderr, "Usage: %s filesize outfile [url ...]\n", argv[0]);
	exit(1);
    }

    /* Allocate the array of Connections. */
    sources = calloc(MAXSOURCES, sizeof(Connection));
    if (sources == NULL) {
	fprintf(stderr, "Out of memory at %s:%d!\n", __FILE__, __LINE__);
	exit(1);
    }

    giFTInit();

    statefile = malloc(strlen(argv[2]) + 5);
    if (!statefile) {
	fprintf(stderr, "Out of memory allocating state filename.\n");
	exit(1);
    }
    sprintf(statefile, "%s.sta", argv[2]);
    if((statef = fopen(statefile, "r")) != NULL) {
	open_mode = "r+";
    } else {
	open_mode = "w";
    }

    /* Open the outfile. */
    outf = fopen(argv[2], open_mode);
    if (!outf) {
	perror("Could not open file");
	exit(1);
    }

    /* Create the DownloadContext. */
    dc = downloadNewContext(filesize, outf, sources, MAXSOURCES);

    if (statef) {
	while(fscanf(statef, "%lu-%lu", &start, &end) == 2) {
	    prev_recv = downloadSetFilled(dc, start, end);
	}
	fclose(statef);
    }

    /* Add the command-line URLs, if any, to the list of sources. */
    for(i=3;i<argc;++i) {
	addsource(dc, argv[i]);
    }

    /* Set the initial time. */
    now = time(NULL);

    /* Main loop */
    while(1) {
	struct timeval tv;
	fd_set rd, wr, ex;
	int nfd = 0;
	int res;
	DownloadStats dstats;
	size_t numselected;

	/* Print periodic status updates. */
	tv.tv_sec = marktime > now ? marktime-now : 0;
	tv.tv_usec = 0;

	FD_ZERO(&rd);
	FD_ZERO(&wr);
	FD_ZERO(&ex);
	if (stdin_open) {
	    FD_SET(0, &rd);
	    nfd = 1;
	}

	/* Let the giFT library figure out which fds need selecting. */
	numselected = selectSetupConnections(sources, MAXSOURCES,
		&nfd, &rd, &wr, &ex, &tv);
	if (!stdin_open && numselected == 0) {
	    /* We're going nowhere. */
	    break;
	}

	res = select(nfd, &rd, &wr, &ex, &tv);

	selectHandleConnections(sources, MAXSOURCES, res, &rd, &wr, &ex,
		NULL, NULL, NULL, NULL, cchandler, &nhosts,
		NULL, NULL, NULL, downloadHandler, dh_context);

	if (res > 0) {
	    if (stdin_open && FD_ISSET(0, &rd)) {
		rcontext = read_and_add_url(0, dc, rcontext);
		if (rcontext == NULL) {
		    stdin_open = 0;
		    fclose(stdin);
		}
	    }
	}

	/* Are we done? */
	if (downloadIsDone(dc)) break;

	/* See if we need to print a status report. */
	now = time(NULL);
	if (now >= marktime) {
	    downloadGetStats(dc, &dstats);
	    printf("Bytes received: %u / %u (%u%%) @ %.2f KB/s; %u hosts\n",
		    dstats.recv, dstats.filesize, dstats.percent,
		    (dstats.recv - prev_recv) / (MARK_INTERVAL * 1024.0),
		    nhosts);
	    saveIntervals(statefile, dc);
	    marktime = now + MARK_INTERVAL;
	    prev_recv = dstats.recv;
	}
    }

    if (downloadIsDone(dc)) {
	printf("Done.\n");
	success = 1;
    } else {
	printf("Failed.\n");
    }
    saveIntervals(statefile, dc);

    /* Close up. */
    free(statefile);
    downloadFreeContext(&dc);
    read_and_add_url(-1, NULL, rcontext);
    fclose(outf);

    return success ? 0 : 1;
}
