#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>
#include <time.h>
#include "giFTpacket.h"
#include "giFTproto.h"
#include "giFTshare.h"
#include "main.h"
#include "nodes.h"

/* The size of the array of outgoing (to node or supernode) Connections */
#define MAXCONNECTIONS 100

/* Don't go out of our way to keep more than this number of the above
 * Connections fully connected and active */
#define MAXACTIVE 20

/* The size of the array of incoming (from clients) Connections */
#define MAXCLIENTS 20

/* The size of the array of incoming (from peers) Connections */
#define MAXPEERS 20

/* The default pathname to the nodes file */
#define NODES_FILE ".giFTnodes"

/* The username we will use is randomly determined.  The networkname
 * (aka software name) we will use is "giFT".  Feel free to change these
 * values if you'd like to masquerade as someone else. */
static char USERNAME[20];
#define NETWORKNAME "giFT"

/* The size of the waiting list of nodes to be contacted. */
#define WAITLIST_SIZE 1000

/* The number of seconds between performing periodic functions such as
 * saving your set of active supernodes. */
#define MARK_INTERVAL 60

/* The port to listen on for incoming connections from giFT clients. */
#define CLIENT_PORT 1213

/* Set this to 1 (the default) if clients can only connect to the above
 * port on localhost (127.0.0.1), which usually means they have to be
 * running on localhost to begin with.  If you set this to 0, any host
 * on the Internet will be able to connect to us as a client, so you
 * will want to provide some protection via a firewall. */
#define CLIENT_LOCALHOST 1

/* The port to listen on for incoming connections from peers. */
#define SERVER_PORT 1214

static struct sockaddr_in SERVER_ADDR;

/**
 * The context structure for ConnectionChangedHandlers.  c is the array
 * of Connections, and restock is a flag indicating whether to try to
 * replace newly free entries with attempts from the waiting list.
 * The shared files referenced by the ShareHandle are send to supernodes
 * that enter the SUPERNODE_CONNECTED state.
 */
typedef struct {
    Connection *c;
    int restock;
    ShareHandle sh;
} CCContext;

/**
 * The context structure for PacketHandlers.  cs is the array of
 * outgoing Connections, and clients is the array of incoming
 * Connections.  The allowPrivateIP flag indicates whether search
 * results in the private IP (RFC 1918) address ranges should be
 * displayed.
 */
typedef struct {
    Connection *cs;
    Connection *clients;
    int allowPrivateIP;
} PHContext;

/**
 * The context structure for QueryHandlers.  This happens to be the same
 * as for PacketHandlers at the moment.
 */
typedef struct {
    Connection *cs;
    Connection *clients;
} QHContext;

/**
 * The global waiting list for peer addresses.  When we find a potential
 * peer address that we're not already trying to connected to, but we're
 * also connected to too many peers already, that address will be put
 * into the waiting list, for later consideration.
 */
static struct sockaddr_in waitlist[WAITLIST_SIZE];
static int waitlist_head = 0;
static int waitlist_len = 0;

/**
 * Add a copy of the given entry to the end of the waiting list, if
 * there's room.
 */
static void waitlist_add(struct sockaddr_in *addr)
{
    int tail = (waitlist_head + waitlist_len) % WAITLIST_SIZE;
    if (waitlist_len == WAITLIST_SIZE) return;
    waitlist[tail] = *addr;
    ++waitlist_len;
}

/**
 * Get a pointer to the next entry in the waiting list (which should be
 * copied from right away), or NULL if the list is empty.
 */
static struct sockaddr_in *waitlist_get()
{
    struct sockaddr_in *retval;

    if (waitlist_len == 0) return NULL;
    retval = &(waitlist[waitlist_head]);
    ++waitlist_head;
    --waitlist_len;
    return retval;
}

/**
 * This AddressHandler is executed once for each address that appears in
 * a packet of peer addresses (types 0x00 and 0x02 from a supernode).
 * It attempts to connect to the given address, and sticks it in the
 * waiting list if we're already connected to too many nodes.
 */
static int addaddress(int which, struct sockaddr *addr, socklen_t addrlen,
	void *context)
{
    Connection *cs = context;
    int res = addToConnectionList(cs, MAXCONNECTIONS, MAXACTIVE,
	    addr, addrlen, USERNAME, NETWORKNAME, &SERVER_ADDR);
    if (res == -1 && addr->sa_family == AF_INET) {
	waitlist_add((struct sockaddr_in *)addr);
    }
    return 0;
}

/**
 * Write an ordinary string using the supplied ClientWriter.
 */
static void write_string(ClientWriter cw, void *cw_context, const char *s)
{
    cw(s, strlen(s), cw_context);
}

/**
 * Write a string using the supplied ClientWriter, but ensure the result
 * is printable.  Use HTML-style escaping.
 */
static void write_escaped_string(ClientWriter cw, void *cw_context,
	const char *s)
{
    while(1) {
	/* Find the longest printable prefix */
	const char *nonprint;
	byte c;
	for(nonprint = s; *nonprint >= ' ' && *nonprint <= '~' &&
		!strchr("&<>\"", *nonprint); ++nonprint) ;
	if (nonprint > s) {
	    cw(s, nonprint-s, cw_context);
	}
	s = nonprint;
	/* Now handle the first nonprintable */
	c = *s;
	if (c == '\0') break;
	if (c == '&') {
	    write_string(cw, cw_context, "&amp;");
	} else if (c == '<') {
	    write_string(cw, cw_context, "&lt;");
	} else if (c == '>') {
	    write_string(cw, cw_context, "&gt;");
	} else if (c == '"') {
	    write_string(cw, cw_context, "&quot;");
	} else {
	    char asstr[10];
	    sprintf(asstr, "&#%u;", c);
	    write_string(cw, cw_context, asstr);
	}
	++s;
    }
}

/**
 * Write a string using the supplied ClientWriter, but ensure the result
 * is printable.  Use URL-style escaping.
 */
static void write_url_escaped_string(ClientWriter cw, void *cw_context,
	const char *s)
{
    while(1) {
	/* Find the longest printable prefix */
	const char *nonprint;
	byte c;
	for(nonprint = s; *nonprint > ' ' && *nonprint <= '~' &&
		!strchr("%+/\\[]()<>'\" ", *nonprint); ++nonprint) ;
	if (nonprint > s) {
	    cw(s, nonprint-s, cw_context);
	}
	s = nonprint;
	/* Now handle the first nonprintable */
	c = *s;
	if (c == '\0') break;
	if (c == ' ') {
	    write_string(cw, cw_context, "+");
	} else {
	    char asstr[10];
	    sprintf(asstr, "%%%02x", c);
	    write_string(cw, cw_context, asstr);
	}
	++s;
    }
}

/**
 * Write a data buffer using the supplied ClientWriter.  The output will
 * be a string of hex digits twice as long as the original buffer.
 */
static void write_hex(ClientWriter cw, void *cw_context, byte *d, size_t l)
{
    static const char hexstr[17] = { "0123456789abcdef" };
    while(l > 0) {
	char hex[2];
	hex[0] = hexstr[(*d)>>4];
	hex[1] = hexstr[(*d)&0x0f];
	cw(hex, 2, cw_context);
	++d;
	--l;
    }
}

/**
 * Write a signed integer using the supplied ClientWriter.
 */
static void write_int(ClientWriter cw, void *cw_context, int v)
{
    char asstr[20];  /* More than big enough to hold the biggest int,
			including a - sign and a trailing NUL. */
    sprintf(asstr, "%d", v);
    write_string(cw, cw_context, asstr);
}

/**
 * Write a number of seconds as MM:SS or HHH:MM:SS using the supplied
 * ClientWriter.
 */
static void write_duration(ClientWriter cw, void *cw_context, int v)
{
    char asstr[30];  /* More than big enough to hold the biggest duration,
			including a - sign and a trailing NUL. */
    if (v >= 3600) {
	sprintf(asstr, "%d:%02d:%02d", v/3600, (v/60)%60, v%60);
    } else {
	sprintf(asstr, "%d:%02d", v/60, v%60);
    }
    write_string(cw, cw_context, asstr);
}

/**
 * Write an unsigned integer in hex, with at least mindigits digits,
 * using the supplied ClientWriter.
 */
static void write_hexint(ClientWriter cw, void *cw_context, unsigned int v,
	int mindigits)
{
    char asstr[20];  /* More than big enough to hold the biggest int,
			including a trailing NUL. */
    if (mindigits > 8) mindigits = 8;
    sprintf(asstr, "%0*x", mindigits, v);
    write_string(cw, cw_context, asstr);
}

/**
 * Decode a multibyte integer from a data buffer.  The pointers to the
 * current position in the data buffer and the length remaining are
 * updated to reflect the data read.  The unsigned integer decoded is
 * returned.  The encoding method used is described in the README file.
 */
static unsigned int decode_int(byte **dp, size_t *lp)
{
    unsigned int v = 0;
    while((*lp)>0) {
	byte n = **dp;
	v <<= 7;
	v |= (n&0x7f);
	++(*dp); --(*lp);
	if (n < 0x80) break;
    }
    return v;
}

/**
 * When we receive a QueryResponse from a supernode, format its
 * contents, and send the result to the client who asked for it.
 */
static void printQueryResponse(QueryResponse qr, Connection *clients,
	int allowPrivateIP)
{
    size_t m, t;
    ClientWriter cw = NULL;
    void *cw_context = NULL;

    if (qr == NULL) return;

    /* Get a ClientWriter which will write to the client which asked for
     * this QueryResponse. */
    if (getClientWriterByQid(clients, MAXCLIENTS, qr->id, &cw,
		&cw_context) < 0) {
	return;
    }

    printf("Response from supernode %s:%u to query id 0x%04x (%u matches)\n",
	    inet_ntoa(qr->addr.sin_addr), ntohs(qr->addr.sin_port),
	    qr->id, qr->numresp);

    /* Write out a formatted version of each file reported using the
     * ClientWriter. */
    for(m=0;m<qr->numresp;++m) {
	FileEntry *fe = &(qr->files[m]);
	if (fe->valid != 1) continue;

	/* See if we want to reject this entry because it's in
	 * private IP space (RFC 1918). */
	if (!allowPrivateIP) {
	    unsigned long addr = ntohl(fe->host.sin_addr.s_addr);
	    if ((addr & 0xff000000) == 0x0a000000 ||
		    (addr & 0xfff00000) == 0xac100000 ||
		    (addr & 0xffff0000) == 0xc0a80000) {
		/* Skip this one. */
		continue;
	    }
	}

	write_string(cw, cw_context, "<ITEM USER=\"");
	write_escaped_string(cw, cw_context,
		fe->username ? fe->username : "???");
	write_string(cw, cw_context, "@");
	write_escaped_string(cw, cw_context,
		fe->networkname ? fe->networkname : "???");
	write_string(cw, cw_context, "\" BANDWIDTHTAG=\"");
	write_hex(cw, cw_context, &(fe->bandwidthtag), 1);
	write_string(cw, cw_context, "\" HASH=\"");
	write_hex(cw, cw_context, fe->d.hash, 20);
	write_string(cw, cw_context, "\" FILESIZE=\"");
	write_int(cw, cw_context, fe->d.filesize);

	for(t=0;t<fe->d.ntags;++t) {
	    byte *d = fe->d.tags[t].data;
	    size_t l = fe->d.tags[t].len;
	    unsigned int v = 0;

	    if (l > 0 && d == NULL) continue;
	    switch(fe->d.tags[t].tag) {
		case FILE_TAG_YEAR:
		    write_string(cw, cw_context, "\" YEAR=\"");
		    v = decode_int(&d, &l);
		    write_int(cw, cw_context, v);
		    break;

		case FILE_TAG_HREF:
		    write_string(cw, cw_context, "\" HREF=\"http://");
		    write_string(cw, cw_context, inet_ntoa(fe->host.sin_addr));
		    write_string(cw, cw_context, ":");
		    write_int(cw, cw_context, fe->host.sin_port ?
			    ntohs(fe->host.sin_port): 1214);
		    write_string(cw, cw_context, "/");
		    write_int(cw, cw_context, fe->d.checksum);
		    write_string(cw, cw_context, "/");
		    write_url_escaped_string(cw, cw_context, d);
		    break;

		case FILE_TAG_TITLE:
		    write_string(cw, cw_context, "\" TITLE=\"");
		    write_escaped_string(cw, cw_context, d);
		    break;

		case FILE_TAG_TIME:
		    write_string(cw, cw_context, "\" TIME=\"");
		    v = decode_int(&d, &l);
		    write_duration(cw, cw_context, v);
		    break;

		case FILE_TAG_ARTIST:
		    write_string(cw, cw_context, "\" ARTIST=\"");
		    write_escaped_string(cw, cw_context, d);
		    break;

		case FILE_TAG_ALBUM:
		    write_string(cw, cw_context, "\" ALBUM=\"");
		    write_escaped_string(cw, cw_context, d);
		    break;

		case FILE_TAG_LANGUAGE:
		    write_string(cw, cw_context, "\" LANGUAGE=\"");
		    write_escaped_string(cw, cw_context, d);
		    break;

		case FILE_TAG_KEYWORDS:
		    write_string(cw, cw_context, "\" KEYWORDS=\"");
		    write_escaped_string(cw, cw_context, d);
		    break;

		case FILE_TAG_RESOLUTION:
		    write_string(cw, cw_context, "\" RESOLUTION=\"");
		    v = decode_int(&d, &l);
		    write_int(cw, cw_context, v);
		    write_string(cw, cw_context, "x");
		    v = decode_int(&d, &l);
		    write_int(cw, cw_context, v);
		    break;

		case FILE_TAG_GENRE:
		    write_string(cw, cw_context, "\" GENRE=\"");
		    write_escaped_string(cw, cw_context, d);
		    break;

		case FILE_TAG_BITDEPTH:
		    write_string(cw, cw_context, "\" BITDEPTH=\"");
		    v = decode_int(&d, &l);
		    write_int(cw, cw_context, v);
		    write_string(cw, cw_context, " bpp");
		    break;

		case 0x12:
		    write_string(cw, cw_context, "\" UNKNOWNTAG12=\"");
		    write_escaped_string(cw, cw_context, d);
		    break;

		case FILE_TAG_QUALITY:
		    write_string(cw, cw_context, "\" BITRATE=\"");
		    v = decode_int(&d, &l);
		    write_int(cw, cw_context, v);
		    write_string(cw, cw_context, " kbps");
		    break;

		case FILE_TAG_VERSION:
		    write_string(cw, cw_context, "\" VERSION=\"");
		    write_escaped_string(cw, cw_context, d);
		    break;

		case FILE_TAG_COMMENT:
		    write_string(cw, cw_context, "\" COMMENT=\"");
		    write_escaped_string(cw, cw_context, d);
		    break;

		case 0x1c:
		    write_string(cw, cw_context, "\" UNKNOWNTAG1c=\"");
		    write_escaped_string(cw, cw_context, d);
		    break;

		case FILE_TAG_RATING:
		    write_string(cw, cw_context, "\" RATING=\"");
		    v = decode_int(&d, &l);
		    switch(v) {
			case 0:
			    write_string(cw, cw_context, "Very poor");
			    break;
			case 1:
			    write_string(cw, cw_context, "Poor");
			    break;
			case 2:
			    write_string(cw, cw_context, "OK");
			    break;
			case 3:
			    write_string(cw, cw_context, "Good");
			    break;
			case 4:
			    write_string(cw, cw_context, "Excellent");
			    break;
			default:
			    write_int(cw, cw_context, v);
			    break;
		    }
		    break;

		default:
		    write_string(cw, cw_context, "\" UNKNOWNTAG");
		    write_hexint(cw, cw_context, fe->d.tags[t].tag, 2);
		    write_string(cw, cw_context, "=\"0x");
		    write_hex(cw, cw_context, d, l);
		    break;
	    }
	}
	write_string(cw, cw_context, "\"/>\n");
    }

    freeCWContext(&cw_context);
}

/**
 * Write statistics about current ConnectionStates to the console.
 */
static void showstats(Connection *cs, Connection *clients, Connection *peers)
{
    int stats[NumConnectionState];
    int i;
    for(i=0;i<NumConnectionState;++i) stats[i] = 0;
    accumulateStats(cs, MAXCONNECTIONS, stats);
    accumulateStats(clients, MAXCLIENTS, stats);
    accumulateStats(peers, MAXPEERS, stats);
    for(i=0;i<NumConnectionState;++i) {
	printf("%-20s: %3d\n", connectionStateName[i], stats[i]);
    }
    printf("(%d waiting)\n\n", waitlist_len);
}

/**
 * This ConnectionChangedHandler is called whenever a Connection has
 * just changed its ConnectionState.  We log a message to the console
 * and, if we now have a free slot, try to connect to the next address
 * on the waiting list.
 */
static int cchandler(int which, ConnectionState from, ConnectionState to,
	void *context)
{
    Connection *cs = ((CCContext *)context)->c;
    ShareHandle sharehandle = ((CCContext *)context)->sh;
    printf("Connection %d changed from %s to %s.\n", which,
	    connectionStateName[from], connectionStateName[to]);
    if (to == CONNECTED_NULL && ((CCContext *)context)->restock) {
	struct sockaddr_in sin, *sinp;
	sinp = waitlist_get();
	if (sinp) {
	    int res;
	    sin = *sinp;
	    res = addToConnectionList(cs, MAXCONNECTIONS, MAXACTIVE,
		    (struct sockaddr *)&sin, sizeof(sin),
		    USERNAME, NETWORKNAME, &SERVER_ADDR);
	    if (res == -1) {
		waitlist_add(&sin);
	    }
	}
    } else if (to == CONNECTED_SUPERNODE_CONNECTED) {
	/* We've just added a new supernode.  Send it the list of files
	 * we're sharing. */
	shareSendAll(sharehandle, cs[which]);
    }

    return 0;
}

/**
 * This PacketHandler is called whenever the low level socket routines
 * have read and decrypted a complete packet.
 */
static int packethandler(int which, int type, size_t size, byte *buf,
	void *context)
{
    PHContext *phc = (PHContext *)context;
    Connection *cs;
    Connection *clients;
    CCContext ccc;

    if (phc == NULL) return -1;
    cs = phc->cs;
    clients = phc->clients;

    /* Log the packet to the console */
    printf("Packet received from %d: type %02x, length %u\n", which,
	    type, size);

    /* Now deal with the packet types we know how to handle. */
    if (type == 0x02 || type == 0x00) {
	readNodeList(type, size, buf, addaddress, cs);
    } else if (type == 0x07) {
	QueryResponse qr = readQueryResponse(type, size, buf);
	if (qr != NULL) {
	    printQueryResponse(qr, clients, phc->allowPrivateIP);
	    freeQueryResponse(&qr);
	}
    } else if (type == 0x08) {
	unsigned short id = readQueryTerminated(type, size, buf);
	ccc.c = clients;
	ccc.restock = 0;
	ccc.sh = NULL;
	flushClientConnection(clients, MAXCLIENTS, id, cchandler, &ccc);
	printf("No more results for query id 0x%04x.\n", id);
    } else if (type == 0x09) {
	SupernodeStats ss;
	if (reportStatsFromConnection(cs[which], &ss) >= 0) {
	    int i;
	    printf("\n%d users sharing files:\n", ss.numUsers);
	    for (i=0;i<NumStatsFileType;++i) {
		printf("%10s: %10d files, %10d GB\n",
			statsFileTypeName[i], ss.numFiles[i],
			ss.numGBytes[i]);
	    }
	    printf("%10s: %10d files, %10d GB\n\n", "Total", ss.numTotFiles,
		    ss.numTotGBytes);
	}
    } else if (type == 0x1d) {
	/* 0x1d indicates the type of software (networkname) the
	 * supernode we're talking to is running.  Just ignore these
	 * packets. */
    } else {
	/* Dump the packet contents to the console. */
	int i,j;
	for(i=0;i<size;++i) {
	    if (i % 16 == 0) printf("\n%04x: ", i);
	    printf("%02x", buf[i]);
	    if (i == size-1 && i % 16 < 15) printf("%*s", 2*(15-(i%16)), "");
	    if (i == size-1 || i % 16 == 15) {
		putchar(' ');
		putchar(' ');
		for(j = i - (i%16); j <= i; ++j) {
		    byte c = buf[j];
		    if (c >= ' ' && c <= '~') putchar(c);
		    else putchar('.');
		}
	    }
	}
	printf("\n\n");
    }
    return 0;
}

/**
 * Unescape the HTML-style escaped string.  This operation is done
 * in-place.  Note that not all of the standard HTML entities are
 * recognized.  Only "&amp;", "&lt;", "&gt;", "&quot;", "&nbsp;",
 * and numerical entites like "&#215;" are converted.
 */
static void unescape_string(char *s)
{
    char *d = s;
    while(*s) {
	char *end, *entname;
	if (*s != '&') {
	    *d++ = *s++;
	    continue;
	}
	end = strchr(s, ';');
	if (!end) {
	    *d++ = *s++;
	    continue;
	}
	entname = s+1;
	*end = '\0';
	s = end+1;
	if (!strcmp(entname, "amp")) {
	    *d++ = '&';
	} else if (!strcmp(entname, "lt")) {
	    *d++ = '<';
	} else if (!strcmp(entname, "gt")) {
	    *d++ = '>';
	} else if (!strcmp(entname, "quot")) {
	    *d++ = '\"';
	} else if (!strcmp(entname, "nbsp")) {
	    *d++ = ' ';
	} else if (entname[0] == '#') {
	    *d++ = atoi(entname+1);
	}
    }
    *d = '\0';
}

typedef struct {
    Connection *cs;
    Connection client;
    unsigned short qid;
    byte *query;
    size_t len;
    size_t count;
} QEContext;

static int query_each(int which, struct sockaddr *addr,
	socklen_t addrlen, void *ah_context)
{
    QEContext *qe = ah_context;
    Connection supernode = qe->cs[which];
    printf("Sending query 0x%04x to supernode %d.\n", qe->qid, which);
    /* Call this to increment the reference count */
    setClientID(qe->client, qe->qid);
    sendComplexQuery(supernode, qe->query, qe->len);
    qe->count++;
    return 0;
}

/**
 * Get the query type based on the prefix of the given string.  Return
 * the number of characters taken up by the prefix.
 */
static int get_qtype(char *value, QueryTermType *qtp)
{
    if (value == NULL) return 0;

    switch(value[0]) {
	case '=':
	    *qtp = QUERY_EQUALS;
	    return 1;
	case '~':
	    *qtp = QUERY_APPROX;
	    return 1;
	case '<':
	    *qtp = QUERY_ATMOST;
	    return 1;
	case '>':
	    *qtp = QUERY_ATLEAST;
	    return 1;
	case '*':
	    *qtp = QUERY_SUBSTRING;
	    return 1;
	default:
	    *qtp = QUERY_EQUALS;
	    return 0;
    }
}

/**
 * Parse a value string that looks like "~102400" into a QueryTermType
 * based on the first char, and a multibyte-encoded int.  Return 0 on
 * success, -1 on failure.
 */
static int make_cmp_int(char *value, QueryTermType *qtp, size_t *qlp,
	byte *qdata)
{
    size_t vv;

    if (value == NULL) return -1;
    value += get_qtype(value, qtp);
    vv = atoi(value);
    if (vv < 128) {
	*qlp = 1;
	qdata[0] = vv;
    } else if (vv < 128*128) {
	*qlp = 2;
	qdata[1] = vv & 0x7f;
	vv >>= 7;
	qdata[0] = vv | 0x80;
    } else if (vv < 128*128*128) {
	*qlp = 3;
	qdata[2] = vv & 0x7f;
	vv >>= 7;
	qdata[1] = (vv&0x7f) | 0x80;
	vv >>= 7;
	qdata[0] = vv | 0x80;
    } else if (vv < 128*128*128*128) {
	*qlp = 4;
	qdata[3] = vv & 0x7f;
	vv >>= 7;
	qdata[2] = (vv&0x7f) | 0x80;
	vv >>= 7;
	qdata[1] = (vv&0x7f) | 0x80;
	vv >>= 7;
	qdata[0] = vv | 0x80;
    } else {
	*qlp = 5;
	qdata[4] = vv & 0x7f;
	vv >>= 7;
	qdata[3] = (vv&0x7f) | 0x80;
	vv >>= 7;
	qdata[2] = (vv&0x7f) | 0x80;
	vv >>= 7;
	qdata[1] = (vv&0x7f) | 0x80;
	vv >>= 7;
	qdata[0] = vv | 0x80;
    }
    return 0;
}

/**
 * Turn a character representing a hex nybble into an int.
 */
static int ctoh(char c)
{
    if (c>='0' && c<='9') return c-'0';
    if (c>='a' && c<='f') return c-'a'+10;
    if (c>='A' && c<='F') return c-'A'+10;
    return 0;
}

/**
 * This routine handles search queries from clients.  The buf is a
 * pointer to a NUL-terminated string of space-separated name=value
 * pairs.  value can be quoted with single or double quotes.
 */
static ConnectionState searchqueryhandler(int which, char *buf,
	void *qh_context)
{
    static unsigned short qid = 0;
    ConnectionState retval = CONNECTED_CLOSED;
    Connection *clients = ((QHContext*)qh_context)->clients;
    Connection *cs = ((QHContext*)qh_context)->cs;
    char *name, *value;
    ComplexQuery cq = NULL;
    int numterms = 0;
    QueryTermType qtype;
    size_t qlen;
    byte qdata[10];  /* Big enough to hold any encoded int */
    QEContext qe;

    /* Increment the query id, but make sure it's not zero. */
    ++qid;
    if (qid == 0) ++qid;

    /* Start building the query */
    cq = createComplexQuery(1000, qid);

    /* Parse out the query into a sequence of name=value pairs. */
    while(*buf) {
	char *end, endchar = ' ';

	while(isspace(*(byte*)buf)) ++buf;
	if (*buf == '\0') break;
	name = buf;
	/* Look for the end of the name */
	end = strchr(name, '=');
	if (!end) {
	    goto clean;
	}
	*end = '\0';
	value = end+1;
	/* Find the end of the value.  This is either at the next space,
	 * if the start of the value is not a ' or ", or else at the
	 * matching quote mark if it is. */
	if (*value == '\0') break;
	if (*value == '\'' || *value == '\"') {
	    endchar = *value;
	    ++value;
	}
	end = strchr(value, endchar);
	if (!end) {
	    if (endchar == '\'' || endchar == '\"') goto clean;
	    end = value + strlen(value);
	    buf = end;
	} else {
	    *end = '\0';
	    buf = end+1;
	}
	unescape_string(value);

	/* If the name matches one we care about, store the unescaped
	 * value appropriately. */
	if (!strcasecmp(name, "realm")) {
	    /* Choose from known values */
	    if (!strcasecmp(value, "audio")) {
		cq->realm = QUERY_REALM_AUDIO;
	    } else if (!strcasecmp(value, "video")) {
		cq->realm = QUERY_REALM_VIDEO;
	    } else if (!strcasecmp(value, "images")) {
		cq->realm = QUERY_REALM_IMAGES;
	    } else if (!strcasecmp(value, "documents")) {
		cq->realm = QUERY_REALM_DOCUMENTS;
	    } else if (!strcasecmp(value, "software")) {
		cq->realm = QUERY_REALM_SOFTWARE;
	    } else {
		cq->realm = QUERY_REALM_EVERYTHING;
	    }
	} else if (!strcasecmp(name, "query")) {
	    numterms = buildComplexQuery(cq, QUERY_SUBSTRING, FILE_TAG_ANY,
		    strlen(value), value);
	    if (numterms < 0) goto clean;
	} else if (!strcasecmp(name, "hash")) {
	    byte pvalue[20];
	    int i;
	    size_t vlen = strlen(value)/2;
	    if (vlen > 20) vlen = 20;
	    memset(pvalue, 0xff, 20);
	    for(i=0;i<vlen;++i) {
		pvalue[i] = (ctoh(value[2*i])<<4) | ctoh(value[2*i+1]);
	    }
	    numterms = buildComplexQuery(cq, QUERY_EQUALS, FILE_TAG_HASH,
		    vlen, pvalue);
	    if (numterms < 0) goto clean;
	} else if (!strcasecmp(name, "title")) {
	    numterms = buildComplexQuery(cq, QUERY_SUBSTRING, FILE_TAG_TITLE,
		    strlen(value), value);
	    if (numterms < 0) goto clean;
	} else if (!strcasecmp(name, "artist")) {
	    numterms = buildComplexQuery(cq, QUERY_SUBSTRING, FILE_TAG_ARTIST,
		    strlen(value), value);
	    if (numterms < 0) goto clean;
	} else if (!strcasecmp(name, "album")) {
	    numterms = buildComplexQuery(cq, QUERY_SUBSTRING, FILE_TAG_ALBUM,
		    strlen(value), value);
	    if (numterms < 0) goto clean;
	} else if (!strcasecmp(name, "language")) {
	    numterms = buildComplexQuery(cq, QUERY_EQUALS, FILE_TAG_LANGUAGE,
		    strlen(value), value);
	    if (numterms < 0) goto clean;
	} else if (!strcasecmp(name, "genre")) {
	    numterms = buildComplexQuery(cq, QUERY_EQUALS, FILE_TAG_GENRE,
		    strlen(value), value);
	    if (numterms < 0) goto clean;
	} else if (!strcasecmp(name, "rating")) {
	    if (make_cmp_int(value, &qtype, &qlen, qdata) < 0) goto clean;
	    numterms = buildComplexQuery(cq, qtype, FILE_TAG_RATING,
		    qlen, qdata);
	    if (numterms < 0) goto clean;
	} else if (!strcasecmp(name, "year")) {
	    if (make_cmp_int(value, &qtype, &qlen, qdata) < 0) goto clean;
	    numterms = buildComplexQuery(cq, qtype, FILE_TAG_YEAR,
		    qlen, qdata);
	    if (numterms < 0) goto clean;
	} else if (!strcasecmp(name, "quality")) {
	    if (make_cmp_int(value, &qtype, &qlen, qdata) < 0) goto clean;
	    numterms = buildComplexQuery(cq, qtype, FILE_TAG_QUALITY,
		    qlen, qdata);
	    if (numterms < 0) goto clean;
	} else if (!strcasecmp(name, "size")) {
	    if (make_cmp_int(value, &qtype, &qlen, qdata) < 0) goto clean;
	    numterms = buildComplexQuery(cq, qtype, FILE_TAG_SIZE,
		    qlen, qdata);
	    if (numterms < 0) goto clean;
	} else if (!strncasecmp(name, "int", 3) && strlen(name) == 5) {
	    int tag = (ctoh(name[3])<<4) | ctoh(name[4]);
	    if (make_cmp_int(value, &qtype, &qlen, qdata) < 0) goto clean;
	    numterms = buildComplexQuery(cq, qtype, tag, qlen, qdata);
	    if (numterms < 0) goto clean;
	} else if (!strncasecmp(name, "data", 4) && strlen(name) == 6) {
	    int tag = (ctoh(name[4])<<4) | ctoh(name[5]);
	    int offset = get_qtype(value, &qtype);
	    numterms = buildComplexQuery(cq, qtype, tag, strlen(value)-offset,
		    value+offset);
	    if (numterms < 0) goto clean;
	}
    }

    if (numterms == 0) goto clean;

    qe.cs = cs;
    qe.client = clients[which];
    qe.qid = cq->qid;
    qe.query = compileComplexQuery(cq, &qe.len);
    qe.count = 0;
    getConnecteds(cs, MAXCONNECTIONS, query_each, &qe);
    if (qe.query) free(qe.query);
    if (qe.count == 0) goto clean;

    retval = CONNECTED_CLIENT_WRITING;

clean:
    freeComplexQuery(&cq);
    return retval;
}

/**
 * Context for the stats_each routine.
 */
typedef struct {
    Connection *cs;
    ClientWriter cw;
    void *cw_context;
    size_t count;
} StatsEachContext;

/**
 * Output a line of statistics for a connected supernode.
 */
static int stats_each(int which, struct sockaddr *addr,
	socklen_t addrlen, void *ah_context)
{
    StatsEachContext *se = ah_context;
    SupernodeStats ss;
    ClientWriter cw = se->cw;
    void *cx = se->cw_context;
    int i;

    if (reportStatsFromConnection(se->cs[which], &ss) < 0) {
	return -1;
    }

    if (ss.lastheard == 0) {
	return -1;
    }

    /* Start writing the stats */
    write_string(cw, cx, "<stat ");

    /* Do we have a parseable address? */
    if (addr->sa_family == AF_INET) {
	struct sockaddr_in *sin = (struct sockaddr_in *)addr;
	char astr[45];
	sprintf(astr, "supernode=\"%s:%u\" ",
		inet_ntoa(sin->sin_addr), ntohs(sin->sin_port));
	write_string(cw, cx, astr);
    }

    write_string(cw, cx, "time=\"");
    write_int(cw, cx, ss.lastheard);
    write_string(cw, cx, "\" users=\"");
    write_int(cw, cx, ss.numUsers);
    write_string(cw, cx, "\" TotalFiles=\"");
    write_int(cw, cx, ss.numTotFiles);
    write_string(cw, cx, "\" TotalGB=\"");
    write_int(cw, cx, ss.numTotGBytes);

    for(i=0;i<NumStatsFileType;++i) {
	write_string(cw, cx, "\" ");
	write_string(cw, cx, statsFileTypeName[i]);
	write_string(cw, cx, "Files=\"");
	write_int(cw, cx, ss.numFiles[i]);
	write_string(cw, cx, "\" ");
	write_string(cw, cx, statsFileTypeName[i]);
	write_string(cw, cx, "GB=\"");
	write_int(cw, cx, ss.numGBytes[i]);
    }

    write_string(cw, cx, "\"/>\n");

    se->count++;
    return 0;
}

/**
 * The client requested some statistics.
 */
static ConnectionState statsqueryhandler(int which, char *buf,
	void *qh_context)
{
    Connection *clients = ((QHContext*)qh_context)->clients;
    Connection *cs = ((QHContext*)qh_context)->cs;
    Connection c = clients[which];
    void *cw_context;
    StatsEachContext se;
    ClientWriter cw = NULL;
    
    if (getClientWriterByConnection(c, &cw, &cw_context) < 0) {
	return CONNECTED_CLOSED;
    }

    /* Output a line of statistics for each connected supernode. */
    se.cs = cs;
    se.cw = cw;
    se.cw_context = cw_context;
    se.count = 0;
    getConnecteds(cs, MAXCONNECTIONS, stats_each, &se);

    return CONNECTED_CLIENT_FLUSHING;
}

/**
 * This QueryHandler is called when a client submits a query to the giFT
 * daemon.  The query will start with "<" and end with "/>".  It returns
 * the ConnectionState the client connection should be put in.  This
 * will be one of CONNECTED_CLIENT_WRITING (if we're planning to write
 * more to the client), CONNECTED_CLIENT_FLUSHING (if we've written
 * everything we need), or CONNECTED_CLOSED (if there's an error).
 */
static ConnectionState queryhandler(int which, char *buf, size_t len,
	void *qh_context)
{
    if (len >= 10 && !strncasecmp(buf, "<search ", 8) &&
	    !strncmp(buf+len-2, "/>", 2)) {
	buf[len-2] = '\0';
	buf += 8;
	return searchqueryhandler(which, buf, qh_context);
    }
    
    if (len >= 9 && !strncasecmp(buf, "<stats ", 7) &&
	    !strncmp(buf+len-2, "/>", 2)) {
	buf[len-2] = '\0';
	buf += 7;
	return statsqueryhandler(which, buf, qh_context);
    }
    
    if (len == 8 && !strcasecmp(buf, "<stats/>")) {
	buf[len-2] = '\0';
	buf += 6;
	return statsqueryhandler(which, buf, qh_context);
    }
    
    /* We didn't recognize the query type */
    return CONNECTED_CLOSED;
}

/**
 * Write the current list of Connections in the ConnectionState
 * CONNECTED_SUPERNODE_CONNECTED to the nodes file.  Return
 * the last modification time to the file.
 */
static time_t write_connecteds(const char *nodesFile, Connection *cs)
{
    return nodesWrite(nodesFile, cs, MAXCONNECTIONS);
}

/**
 * Read a potential list of supernodes from the nodes file.
 * Return the number of connections attempted.
 */
static int read_connecteds(const char *nodesFile, Connection *cs,
	time_t last_time)
{
    return nodesRead(nodesFile, cs, MAXCONNECTIONS, MAXACTIVE, USERNAME,
	    NETWORKNAME, &SERVER_ADDR, last_time);
}

static void Usage(char *argv0)
{
    printf("Usage: %s [-a] [-p] [-n nodesfile] [-P minutes] [-s sharedir] "
	    "[-l 10.1.2.3]\n", argv0);
    printf("-a           : Allow search results from the private IP "
	    "(RFC 1918) range\n");
    printf("-p           : Just post the nodes file to the web site, "
	    "then exit.\n");
    printf("-n nodesfile : The location of the nodes file.  Can be a full "
	    "path.\n");
    printf("               Defaults to .giFTnodes.\n");
    printf("-P interval  : Post the nodes to the web site every "
	    "interval minutes.\n");
    printf("               Defaults to 240 (every 4 hours).  0 to "
	    "disable posting.\n");
    printf("-s sharedir  : Specify the directory containing files "
	    "to share.\n");
    printf("               Defaults to \"share\".  \"\" to disable "
	    "sharing.\n");
    printf("-l 10.1.2.3  : Listen for peers on the specified IP address.\n");
    printf("               Defaults to all addresses.\n");
    exit(1);
}

int main(int argc, char **argv)
{
    Connection *c, *clients, *peers;
    PHContext phc;
    QHContext qhc;
    int clientfd;
    int serverfd;
    struct sockaddr_in addr;
    socklen_t addrlen = sizeof(addr);
    int i,res,nfd;
    fd_set rd,wr,ex;
    unsigned int one = 1;
    time_t now = time(NULL);
    time_t marktime = now + MARK_INTERVAL;
    int stdin_open = 1;
    time_t last_giFTnode_time = 0;
    char *argv0 = argv[0];
    int postInterval = 3600*4;
    time_t postTime = 0;
    int allow_probe = 0;
    ShareHandle sharehandle = NULL;
    const char *sharedir = "share";
    time_t shareTime = 0;
    int shareInterval = 60;
    unsigned int server_ip = INADDR_ANY;
    PeerServerContext psc;
    const char *nodesFile = NODES_FILE;
    int justpost = 0;
    int allowPrivateIP = 0;
    
	memset((void*)&addr, (char)0, addrlen);
	
    /* Parse command-line options */
    argc--;
    argv++;
    while(argc > 0) {
	if (!strcmp(*argv, "-p")) {
	    /* Just post the nodes file to the URL, but we should wait
	     * in case -n is specified later. */
	    justpost = 1;
	    argv += 1;
	    argc -= 1;
	    continue;
	}
	if (!strcmp(*argv, "-n") && argc > 1) {
	    nodesFile = argv[1];
	    argv += 2;
	    argc -= 2;
	    continue;
	}
	if (!strcmp(*argv, "-P") && argc > 1) {
	    postInterval = atoi(argv[1]) * 60;
	    argv += 2;
	    argc -= 2;
	    continue;
	}
	if (!strcmp(*argv, "-s") && argc > 1) {
	    sharedir = argv[1];
	    if (sharedir[0] == '\0') {
		shareInterval = 0;
	    }
	    argv += 2;
	    argc -= 2;
	    continue;
	}
	if (!strcmp(*argv, "-l") && argc > 1) {
	    server_ip = inet_addr(argv[1]);
	    argv += 2;
	    argc -= 2;
	    continue;
	}
	if (!strcmp(*argv, "-a")) {
	    /* Allow search results from the private IP (RFC 1918) range. */
	    allowPrivateIP = 1;
	    argv += 1;
	    argc -= 1;
	    continue;
	}
	Usage(argv0);
    }

    /* See if we're only meant to just post and exit. */
    if (justpost) {
	nodesDoPost(nodesFile);
	exit(0);
    }

    /* Fill in the SERVER_ADDR struct */
	memset((void*)&SERVER_ADDR, (char)0, sizeof(struct sockaddr_in));
    SERVER_ADDR.sin_family = AF_INET;
    SERVER_ADDR.sin_port = htons(SERVER_PORT);
    SERVER_ADDR.sin_addr.s_addr = server_ip;

    /* Initialize the giFT library */
    giFTInit();

    /* Seed the PRNG used to create random usernames and pick random
     * supernodes.  This doesn't need to be any better than what the
     * usual C library gives us. */
    srand(time(NULL));

    /* Allocate arrays of Connections for outgoing and incoming
     * Connections */
    c = calloc(MAXCONNECTIONS, sizeof(Connection));
    clients = calloc(MAXCLIENTS, sizeof(Connection));
    peers = calloc(MAXPEERS, sizeof(Connection));
    if (c == NULL || clients == NULL || peers == NULL) {
	fprintf(stderr, "Out of memory at %s:%d\n", __FILE__, __LINE__);
	exit(1);
    }

    /* Set up the PacketHandler and QueryHandler contexts */
    phc.cs = c;
    phc.clients = clients;
    phc.allowPrivateIP = allowPrivateIP;
    qhc.cs = c;
    qhc.clients = clients;

    /* Listen for giFT clients */
    clientfd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (clientfd < 0) {
	perror("socket");
	exit(1);
    }
    setsockopt(clientfd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));

    addr.sin_family = AF_INET;
    addr.sin_port = htons(CLIENT_PORT);
    addr.sin_addr.s_addr =
	inet_addr(CLIENT_LOCALHOST ? "127.0.0.1" : "0.0.0.0");
    if (bind(clientfd, (struct sockaddr *)&addr, addrlen) < 0) {
	perror("bind");
	exit(1);
    }
    if (listen(clientfd, 5) < 0) {
	perror("listen");
	exit(1);
    }

    /* Listen for connections from other nodes */
    serverfd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
    if (clientfd < 0) {
	perror("socket");
	exit(1);
    }
    setsockopt(serverfd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));

    addr = SERVER_ADDR;
    if (bind(serverfd, (struct sockaddr *)&addr, addrlen) < 0) {
	perror("bind");
	exit(1);
    }
    if (listen(serverfd, 5) < 0) {
	perror("listen");
	exit(1);
    }

    /* Get the list of files to share. */
    sharehandle = shareNewHandle(sharedir);
    if (shareInterval > 0) {
	shareUpdate(sharehandle, c, MAXCONNECTIONS);
    }

    /* Read nodes from the nodes file */
    read_connecteds(nodesFile, c, 0);

    /* Make up a random username */
    sprintf(USERNAME, "%08lx%08x", now, rand());

    /* Set up the PeerServerContext. */
    psc.sh = sharehandle;
    psc.server_addr = SERVER_ADDR;
    psc.username = USERNAME;
    psc.networkname = NETWORKNAME;
    psc.cs = c;
    psc.numcs = MAXCONNECTIONS;
    psc.peers = peers;
    psc.numpeers = MAXPEERS;

    /* Main loop */
    while(1) {
	size_t numselected;
	CCContext ccc;
	struct timeval tv;
	now = time(NULL);

	/* How long until we need to perform our periodic functions? */
	tv.tv_sec = marktime > now ? marktime-now : 0;
	tv.tv_usec = 0;

	/* If stdin is still open, listen on it.  Also listen for
	 * incoming client connections. */
	FD_ZERO(&rd);
	FD_ZERO(&wr);
	FD_ZERO(&ex);
	if (stdin_open) FD_SET(0, &rd);
	FD_SET(clientfd, &rd);
	nfd = clientfd + 1;
	FD_SET(serverfd, &rd);
	if (serverfd + 1 > nfd) nfd = serverfd + 1;

	/* Allow the giFT library to update the select parameters
	 * according to the state of the various Connections. */
	numselected = selectSetupConnections(c, MAXCONNECTIONS, &nfd,
		&rd, &wr, &ex, &tv);
	selectSetupConnections(clients, MAXCLIENTS, &nfd, &rd, &wr, &ex, &tv);
	selectSetupConnections(peers, MAXPEERS, &nfd, &rd, &wr, &ex, &tv);
	if (numselected == 0) {
	    /* All our connections are dead.  Just do some cleanup. */
	    for(i=0;i<MAXCONNECTIONS;++i) {
		closeConnection(c+i);
	    }

	    /* Fetch a nodes file from the Internet */
	    printf("No supernodes to talk to... attempting to fetch "
		    "nodes file.\n");
	    nodesBlockingGet(nodesFile);
	    last_giFTnode_time = 0;
	    if (read_connecteds(nodesFile, c, last_giFTnode_time) > 0) {
		continue;
	    }

	    if (allow_probe) {
		printf("Still no supernodes to talk to.  "
			"Probing 24.0.0.0/8...\n");
		/* Start over with some random sites from the 24.0.0.0/8
		 * network. */
		for(i=0;i<MAXCONNECTIONS;++i) {
		    struct sockaddr_in sin;
		    socklen_t sinlen = sizeof(sin);

		    sin.sin_family = AF_INET;
		    sin.sin_port = htons(1214);
		    sin.sin_addr.s_addr = htonl( 24<<24 |
			    ((rand()&0xff)<<16) | 
			    ((rand()&0xff)<<8) | 
			    ((rand()&0xff)));
		    res = addToConnectionList(c, MAXCONNECTIONS, MAXACTIVE,
			    (struct sockaddr *)&sin, sinlen,
			    USERNAME, NETWORKNAME, &SERVER_ADDR);
		    if (res == -1) {
			waitlist_add(&sin);
		    }
		}
		continue;
	    }
	}

	/* Do the select.  This should be the only place in the entire
	 * daemon where we block, except for things like fetching the
	 * nodes file from the web, when we don't have anything else
	 * to do, anyway.  It's possible that simply reading or writing
	 * the nodes file may block, but we'll ignore that issue. */
	res = select(nfd, &rd, &wr, &ex, &tv);

	/* Handle the results of the select. */
	ccc.c = c;
	ccc.restock = 1;
	ccc.sh = sharehandle;
	selectHandleConnections(c, MAXCONNECTIONS, res, &rd, &wr, &ex,
		packethandler, &phc, NULL, NULL, cchandler, &ccc,
		addaddress, c, NULL, NULL, NULL);
	ccc.c = clients;
	ccc.restock = 0;
	ccc.sh = NULL;
	selectHandleConnections(clients, MAXCLIENTS, res, &rd, &wr, &ex,
		NULL, NULL, queryhandler, &qhc, cchandler, &ccc,
		NULL, NULL, NULL, NULL, NULL);
	ccc.c = peers;
	ccc.restock = 0;
	ccc.sh = NULL;
	selectHandleConnections(peers, MAXPEERS, res, &rd, &wr, &ex,
		NULL, NULL, NULL, NULL, cchandler, &ccc,
		NULL, NULL, &psc, NULL, NULL);

	/* Also check for input from the console, and new clients
	 * or peers connecting to us. */
	if (res > 0) {
	    if (stdin_open && FD_ISSET(0, &rd)) {
		char buf[1024];
		char *nl;
		if (fgets(buf, 1024, stdin) == NULL) {
		    /* Our stdin saw EOF.  Close it and never check it
		     * again. */
		    stdin_open = 0;
		    fclose(stdin);
		} else {
		    /* Read a command from the console.  We only know a
		     * couple of commands right now. */
		    nl = strchr(buf, '\n');
		    if (nl) *nl = '\0';
		    if (!strcmp(buf, "quit")) {
			break;
		    } else if (!strcmp(buf, "stats")) {
			showstats(c, clients, peers);
		    } else if (!strcmp(buf, "post")) {
			nodesPost(nodesFile);
		    } else {
			printf("Unknown command \"%s\".\n", buf);
		    }
		}
	    }
	    if (FD_ISSET(clientfd, &rd)) {
		/* Try to accept a new client connection. */
		struct sockaddr addr;
		socklen_t addrlen = sizeof(addr);
		int newfd = accept(clientfd, &addr, &addrlen);
		if (newfd >= 0) {
		    res = addToClientsList(clients, MAXCLIENTS, &addr,
			    sizeof(addr), newfd);
		    if (res == -1) {
			/* We can't deal with this one just now because
			 * our array of client Connections is full.  Get
			 * rid of it. */
			close(newfd);
		    }
		}
	    }
	    if (FD_ISSET(serverfd, &rd)) {
		/* Try to accept a new peer connection. */
		struct sockaddr addr;
		socklen_t addrlen = sizeof(addr);
		int newfd = accept(serverfd, &addr, &addrlen);
		if (newfd >= 0) {
		    res = addToPeersList(peers, MAXPEERS, &addr,
			    sizeof(addr), newfd);
		    if (res == -1) {
			/* We can't deal with this one just now because
			 * our array of peer Connections is full.  Get
			 * rid of it. */
			close(newfd);
		    }
		}
	    }
	}
	/* See if we need to mark time */
	now = time(NULL);
	if (now >= marktime) {
	    /* Read the connected supernodes, in case the file has
	     * changed recently. */
	    read_connecteds(nodesFile, c, last_giFTnode_time);

	    /* Save the current list of connected supernodes. */
	    last_giFTnode_time = write_connecteds(nodesFile, c);
	    marktime = now + MARK_INTERVAL;

	    /* Periodically send our list of supernodes to a central
	     * server. */
	    if (postInterval > 0 && now >= postTime) {
		nodesPost(nodesFile);
		postTime = now + postInterval;
	    }

	    /* Periodically send our list of shared files to our
	     * supernodes. */
	    if (shareInterval > 0 && now >= shareTime) {
		shareUpdate(sharehandle, c, MAXCONNECTIONS);
		shareTime = now + shareInterval;
	    }
	}
    }

    /* When we quit, save the list of connected supernodes for next
     * time. */
    write_connecteds(nodesFile, c);

    shareFreeHandle(&sharehandle);

    return 0;
}
